"""
Utilities for converting a neuroglancer video script to a Google Sheet and back.

The neuroglancer video_tool uses a very simple format for video scripts:

.. code-block:: text

    <url>
    <duration>
    <url>
    <duration>
    ...

That format is simple to create, but annoying to edit.

This utility will load a video script into a Google Sheet in which each
link has been converted to a JSON state and written out to a single column,
with subsequent links in subsequent columns, with JSON settings aligned so
that settings are easily compared across a whole row.

You can edit settings in the Google sheet, and then import them back into
a video script using this utility.

Recommended Google Sheet template:
https://docs.google.com/spreadsheets/d/1SC-szBtssuqMnVrNdZv9LHj4UB0rUFMwZaxeDnekA4c/edit?gid=162774670#gid=162774670

Copy that template and pass the URL of your copied sheet to this tool.
Make sure your sheet is visible to "anyone with the link" and editable at least by you
(via your ``GOOGLE_APPLICATION_CREDENTIALS``).

Example usage:

.. code-block:: bash

    SHEET_URL=https://docs.google.com/spreadsheets/d/15Q1oRLZQd1MWZOjRr_DyIkFbbhMZCSP4sbU78JXkoSM/edit?gid=162774670#gid=162774670
    python ngvideo-helper script-to-sheet video_script.txt $SHEET_URL
    python ngvideo-helper sheet-to-script $SHEET_URL video_script.txt

To install the neuroglancer ``video_tool``:

.. code-block:: bash

    pixi init selenium geckodriver pillow numpy requests tornado six google-apitools google-auth atomicwrites ffmpeg
    pixi add --pypi neuroglancer

Note:
    This tool accepts (and produces) a non-standard format for neuroglancer video scripts
    which allows comments (starting with '#') and blank lines.
    When writing to a video script, comments can be excluded with the ``--no-comments`` flag.
    Otherwise, you'll have to exclude them yourself before running the neuroglancer video_tool:

    .. code-block:: bash

        pixi run python -m neuroglancer.tool.video_tool render \\
            --browser firefox --no-headless --refresh-browser-timeout=10 \\
            --hide-axis-lines --height=2160 --width=2160 \\
            <(grep -v '^#' my-video.txt | grep -v '^$') my-video-frames

        pixi run ffmpeg -r 24 -f image2 -s 2160x2160 -i my-video-frames/%07d.png \\
            -vcodec libx264 -crf 25 -pix_fmt yuv420p my-video.mp4
"""
import argparse
import csv
import json
import os
import re
import urllib.parse
import warnings
from collections import namedtuple

import pandas as pd

ScriptItem = namedtuple('ScriptItem', ['comment', 'transition_duration', 'link', 'state', 'link_line_number'])


def main():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    subparsers = parser.add_subparsers(dest='command', required=True)

    # Read command
    read_parser = subparsers.add_parser('script-to-sheet', help='Read video script and upload to a Google spreadsheet')
    read_parser.add_argument('video_script_path')
    read_parser.add_argument('output_sheet', default='', nargs='?')

    # Write command
    write_parser = subparsers.add_parser('sheet-to-script', help='Write from Google Sheet to video script')
    write_parser.add_argument(
        '--no-comments', action='store_true',
        help='Do not include comments in the output script, making it a valid neuroglancer '
             'video script without any postprocessing needed.')
    write_parser.add_argument('google_sheet_url')
    write_parser.add_argument('output_script_path')
    write_parser.add_argument(
        '--ng-server', default='neuroglancer-demo.appspot.com',
        help='Neuroglancer server URL (default: neuroglancer-demo.appspot.com)'
    )

    args = parser.parse_args()

    if args.command == 'script-to-sheet':
        item_df = video_script_to_dataframe(args.video_script_path)
        if args.output_sheet.startswith('http'):
            overwrite_google_sheet(item_df, args.output_sheet)
        else:
            write_tsv(item_df, args.output_sheet)
    elif args.command == 'sheet-to-script':
        load_from_google_sheet(args.google_sheet_url, args.output_script_path, not args.no_comments, args.ng_server)


def video_script_to_dataframe(video_script_path: str):
    print(f"Reading video script from {video_script_path}")
    with open(video_script_path, 'r') as f:
        lines = f.readlines()

    script_items = []
    cur_comment = ''
    cur_duration = 0
    for i, line in enumerate(lines, start=1):
        line = line.strip()
        if line == '':
            continue

        if line.startswith('#'):
            cur_comment += line + '\n'
            continue

        if not line.startswith('http'):
            try:
                duration = float(line)
            except ValueError:
                raise RuntimeError(f"Line {i} is not a comment, link, or duration:\n{line}")  # noqa
            else:
                if cur_duration != 0:
                    raise RuntimeError(f"Line {i} lists a duration after no link.")
                cur_duration = duration
                continue

        state = parse_nglink(line)
        item = ScriptItem(cur_comment, cur_duration, line, state, i)
        script_items.append(item)
        cur_comment = ''
        cur_duration = 0

    nested_keys = [list_nested_keys(item.state) for item in script_items]
    merged_keys = nested_keys[0]
    for nk in nested_keys[1:]:
        merged_keys = merge_nested_keys(merged_keys, nk)

    state_jsons = []
    for item in script_items:
        j = dump_aligned_json(item.state, merged_keys)
        state_jsons.append(j)

    state_series = []
    for j in state_jsons:
        state_series.append(pd.Series(j.splitlines()))

    state_df = pd.DataFrame(state_series)

    item_df = pd.DataFrame({
        "comment": [item.comment for item in script_items],
        "transition_duration": [item.transition_duration for item in script_items],
        "link_line_number": [item.link_line_number for item in script_items],
    })
    item_df = pd.concat([item_df, state_df], axis=1)
    return item_df


def write_tsv(item_df: pd.DataFrame, output_tsv_path: str):
    print(f"Writing TSV to {output_tsv_path}")
    item_df.T.to_csv(
        output_tsv_path,
        header=False,
        index=True,
        sep='\t',
        quoting=csv.QUOTE_ALL
    )


def overwrite_google_sheet(item_df: pd.DataFrame, sheet_url: str):
    """
    Overwrite a Google Sheet with the given DataFrame.
    Requires the gspread package.
    """
    import gspread

    gid = re.search(r'gid=(\d+)', sheet_url).group(1)

    gc = gspread.service_account(os.environ['GOOGLE_APPLICATION_CREDENTIALS'])
    sh = gc.open_by_url(sheet_url)
    ws = sh.get_worksheet_by_id(gid)
    print(f"Overwriting Google sheet '{ws.title}'")
    ws.clear()
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message=".*Method signature's arguments 'range_name' and 'values'.*")
        ws.update(
            values=item_df.T.reset_index().values.tolist(),
            range_name='A1'
        )


def load_from_google_sheet(worksheet_url: str, output_script_path: str, include_comments=True, ng_server: str = 'neuroglancer-demo.appspot.com'):
    if not ng_server.startswith('http'):
        ng_server = f'https://{ng_server}'

    gid = re.search(r'gid=(\d+)', worksheet_url).group(1)

    sheet_url = re.match(r'(.*)/.*?\?.*', worksheet_url).group(1)
    export_url = f'{sheet_url}/export?format=csv&gid={gid}'
    print(f"Loading from Google sheet: '{export_url}'")
    df = pd.read_csv(export_url, header=None, dtype=str).set_index(0).fillna('').T
    df['comment'] = df['comment'].str.strip()
    df['state'] = df.loc[:, '0':].sum(axis=1).map(json.loads)
    df['link'] = df['state'].map(lambda state: format_nglink(ng_server, state))

    print(f"Writing script to {output_script_path}")
    with open(output_script_path, 'w') as f:
        for i, row in enumerate(df.itertuples()):
            if include_comments and row.comment:
                f.write(f'{row.comment}\n')
            if i != 0:
                f.write(f'{row.transition_duration}\n')
            f.write(f'{row.link}\n')
            if include_comments:
                f.write('\n')


def parse_nglink(link):
    _, pseudo_json = link.split('#!')
    pseudo_json = urllib.parse.unquote(pseudo_json)
    data = json.loads(pseudo_json)
    return data


def format_nglink(ng_server, link_json_settings):
    return ng_server + '/#!' + urllib.parse.quote(json.dumps(link_json_settings))


def list_nested_keys(obj):
    keys = []
    _list_nested_keys(obj, (), keys)
    return keys


def _list_nested_keys(obj, base_key, keys):
    match obj:
        case float() | int() | bool() | str():
            keys.append(base_key)
        case list() if all(isinstance(v, (float, int, bool)) for v in obj):
            keys.append(base_key)
        case list() if all(isinstance(v, str) for v in obj) and pd.Series(pd.to_numeric(obj, errors='coerce')).notnull().all():
            keys.append(base_key)
        case list():
            keys.append((*base_key, '['))
            for i, v in enumerate(obj):
                _list_nested_keys(v, (*base_key, i), keys)
            keys.append((*base_key, ']'))
        case dict():
            keys.append((*base_key, '{'))
            for k, v in obj.items():
                _list_nested_keys(v, (*base_key, k), keys)
            keys.append((*base_key, '}'))
        case _:
            raise ValueError(f"Unknown type: {type(obj)}")


def merge_nested_keys(a, b):
    common_keys = set(a) & set(b)
    a_common = list(k for k in a if k in common_keys)
    b_common = list(k for k in b if k in common_keys)
    if a_common != b_common:
        raise ValueError(f"Common keys occur in different orders: {a_common} vs. {b_common}")

    a, b = list(a), list(b)

    new_keys = []
    while a or b:
        while a and (ka := a.pop(0)) not in common_keys:
            new_keys.append(ka)

        while b and (kb := b.pop(0)) not in common_keys:
            new_keys.append(kb)

        if ka == kb:
            new_keys.append(ka)

    return new_keys


def dump_aligned_json(obj, all_nested_keys, indent=2):
    key_offsets = {k: i for i, k in enumerate(all_nested_keys)}
    lines = [''] * len(all_nested_keys)

    match obj:
        case dict():
            key = ('{',)
        case list():
            key = ('[',)
        case _:
            raise ValueError(f"Unknown type: {type(obj)}")

    _dump_aligned_json(obj, key, 0, key_offsets, indent, lines, '')
    return '\n'.join(lines)


def _dump_aligned_json(obj, key, indent_level, key_offsets, indent, lines, trailing_comma):
    padding = " " * indent_level * indent
    match obj:
        case float() | int() | bool():
            line = lines[key_offsets[key]]
            lines[key_offsets[key]] = padding + line + str(obj).lower() + trailing_comma
        case str():
            line = lines[key_offsets[key]]
            lines[key_offsets[key]] = padding + line + f'"{obj}"' + trailing_comma
        case list() if all(isinstance(v, (float, int, bool)) for v in obj):
            line = lines[key_offsets[key]]
            lines[key_offsets[key]] = padding + line + "[" + ", ".join(map(str, obj)) + "]" + trailing_comma
        case list() if all(isinstance(v, str) for v in obj) and pd.Series(pd.to_numeric(obj, errors='coerce')).notnull().all():
            line = lines[key_offsets[key]]
            lines[key_offsets[key]] = padding + line + "[" + ", ".join(f'"{v}"' for v in obj) + "]" + trailing_comma
        case list():
            *key, suffix = key
            assert suffix == '['
            line = lines[key_offsets[(*key, '[')]]
            lines[key_offsets[(*key, '[')]] = padding + line + "["
            for i, v in enumerate(obj):
                match v:
                    case float() | int() | bool() | str():
                        item_key = (*key, i)
                    case list() if all(isinstance(x, (float, int)) for x in v):
                        item_key = (*key, i)
                    case list():
                        item_key = (*key, i, '[')
                    case dict():
                        item_key = (*key, i, '{')
                    case _:
                        raise ValueError(f"Unknown type: {type(v)}")

                tc = ',' if i < len(obj) - 1 else ''
                _dump_aligned_json(v, item_key, indent_level+1, key_offsets, indent, lines, tc)
            lines[key_offsets[(*key, ']')]] += padding + "]" + trailing_comma
            return lines
        case dict():
            *key, suffix = key
            assert suffix == '{'
            line = lines[key_offsets[(*key, '{')]]
            lines[key_offsets[(*key, '{')]] = padding + line + "{"
            for i, (k, v) in enumerate(obj.items()):
                match v:
                    case float() | int() | bool() | str():
                        item_key = (*key, k)
                    case list() if all(isinstance(x, (float, int, bool)) for x in v):
                        item_key = (*key, k)
                    case list() if all(isinstance(x, str) for x in v) and pd.Series(pd.to_numeric(v, errors='coerce')).notnull().all():
                        item_key = (*key, k)
                    case list():
                        item_key = (*key, k, '[')
                    case dict():
                        item_key = (*key, k, '{')
                    case _:
                        raise ValueError(f"Unknown type: {type(v)}")

                lines[key_offsets[item_key]] += padding + f'"{k}": '
                tc = ',' if i < len(obj) - 1 else ''
                _dump_aligned_json(v, item_key, indent_level+1, key_offsets, indent, lines, tc)
            lines[key_offsets[(*key, '}')]] += padding + "}" + trailing_comma
        case _:
            raise ValueError(f"Unknown type: {type(obj)}")


if __name__ == '__main__':
    main()



"""
Basic convenience functions for downloading and parsing neuroglancer states.
"""
import re
import json
import urllib
import logging

logger = logging.getLogger(__name__)


def parse_nglink(link):
    """
    Given a neuroglancer link, return the corresponding JSON state.

    Arg:
        link: str
            A neuroglancer link such as:
            https://neuroglancer-demo.appspot.com/#!gs://flyem-male-cns/v0.9/male-cns-v0.9.json

            Also works with fully expanded neuroglancer links.

    Returns:
        dict
    """
    _, pseudo_json = link.split('#!')
    if pseudo_json.endswith('.json'):
        return download_ngstate(pseudo_json)
    pseudo_json = urllib.parse.unquote(pseudo_json)
    data = json.loads(pseudo_json)
    return data


def download_ngstate(link):
    """
    Given a neuroglancer link which references a JSON state file,
    download the file and return the enclosed JSON data.

    Arg:
        link: str
            A neuroglancer link such as:
            https://neuroglancer-demo.appspot.com/#!gs://flyem-male-cns/v0.9/male-cns-v0.9.json

    Returns:
        dict
    """
    import requests
    if link.startswith('gs://'):
        url = f'https://storage.googleapis.com/{link[len("gs://"):]}'
        return requests.get(url, timeout=10).json()

    if not link.startswith('http'):
        raise ValueError(f"Don't understand state link: {link}")

    if link.count('://') == 1:
        return requests.get(link, timeout=10).json()

    if link.count('://') == 2:
        url = f'https://storage.googleapis.com/{link.split("://")[2]}'
        return requests.get(url, timeout=10).json()

    raise ValueError(f"Don't understand state link: {link}")


def encode_ngstate(ng_server, link_json_settings):
    """
    Produce a fully expanded URL from a neuroglancer server and a JSON state.
    """
    return ng_server + '/#!' + urllib.parse.quote(json.dumps(link_json_settings))


def layer_dict(state):
    """
    Given a neuroglancer JSON state, return a dictionary of the layer states, keyed by layer name.
    """
    return {layer['name']: layer for layer in state['layers']}


def layer_state(state, name_pattern):
    """
    Given a neuroglancer JSON state and a regex pattern,
    return the layer state whose name matches the pattern.
    If more than one layer matches the pattern, raise an error.
    """
    layer = None
    matches = []
    for layer in state['layers']:
        if re.match(name_pattern, layer['name']):
            matches.append(layer)
    if len(matches) > 1:
        matched_names = [match['name'] for match in matches]
        raise RuntimeError(f"Found more than one layer matching to the regex '{name_pattern}': {matched_names}")
    return matches[0]



import numpy as np
from numba import njit, guvectorize


def compressed_morton_code(grid_coord_c_order, grid_shape_c_order):
    """
    Return the compressed morton code[1] for a grid coordinate
    that resides in a grid with the given shape.

    [1]: https://github.com/google/neuroglancer/blob/master/src/datasource/precomputed/volume.md#compressed-morton-code

    This implementation uses numba.guvectorize().
    Multiple coordinates and/or grid shapes can be provided simultaneously,
    and they will be broadcasted together to produce an output of the appropriate shape.

    Important:
    
        This function expects the grid coordinate to be in C order,
        i.e. the fastest-changing stride is listed last.
        For typical 3D arrays, C-order indexing is [z, y, x],
        so pass your grid coordinates here in [z, y, x] order, too.

    Examples:

        In [10]: compressed_morton_code(
            ...:     [1, 5, 3],
            ...:     [3, 16, 5]
            ...: )
        Out[10]: array(143, dtype=uint64)

        In [11]: compressed_morton_code(
            ...:     [[1, 5, 3], [2, 3, 4]],
            ...:     [3, 16, 5]
            ...: )
        Out[11]: array([143, 114], dtype=uint64)

        In [12]: compressed_morton_code(
            ...:     [1, 5, 3],
            ...:     [[3, 16, 5], [6, 32, 10]]
            ...: )
        Out[12]: array([143, 143], dtype=uint64)

        In [13]: compressed_morton_code(
            ...:     [[1, 5, 3], [2, 3, 4]],
            ...:     [[3, 16, 5], [6, 32, 10]]
            ...: )
        Out[13]: array([143, 114], dtype=uint64)
    """
    grid_coord = np.asarray(grid_coord_c_order, np.uint64)
    grid_shape = np.asarray(grid_shape_c_order, np.uint64)
    if (grid_coord >= grid_shape).any():
        raise ValueError(f'grid_coord contains out-of-bounds values relative to grid_shape')

    axis_bits = np.ceil(np.log2(grid_shape)).astype(np.int8)
    output_shape = np.broadcast_shapes(grid_shape.shape, grid_coord.shape)[:-1]
    output_code = np.zeros(output_shape, dtype=np.uint64)
    __compressed_morton_code(grid_coord, axis_bits, output_code)
    return output_code


@guvectorize('(d),(d)->()', nopython=True)
def __compressed_morton_code(grid_coord, axis_bits, result):
    D = len(axis_bits)
    curr_axis_pos = np.zeros(D, dtype=np.uint64)
    curr_axis = 0
    output_pos = np.uint64(0)
    output_code = np.uint64(0)

    for _ in range(D * 64):
        curr_axis = (curr_axis - 1) % D
        if curr_axis_pos[curr_axis] >= axis_bits[curr_axis]:
            continue

        input_pos = curr_axis_pos[curr_axis]
        output_code |= ((grid_coord[curr_axis] >> input_pos) & np.uint64(1)) << output_pos
        output_pos += np.uint64(1)
        curr_axis_pos[curr_axis] += 1

    # With guvectorize, we treat a scalar output as if it had a single element.
    # https://numba.readthedocs.io/en/stable/user/vectorize.html#scalar-return-values
    result[0] = output_code


@njit
def compressed_morton_code_no_broadcast(grid_coord_c_order, grid_shape_c_order):
    """
    Same as compressed_morton_code(), but for when both grid_coord
    and grid_shape are known to have ndim 1 or 2 and both arguments
    have dtype np.uint64, which allows us to easily predict the output
    shape without needing to call np.broadcast_shapes().

    Therefore this whole function can be wrapped with @njit,
    and can be called from within other jit-compiled functions.
    """
    grid_coord = grid_coord_c_order
    grid_shape = grid_shape_c_order

    axis_bits = np.ceil(np.log2(grid_shape)).astype(np.int8)
    if grid_coord.ndim == 1:
        output_code = np.zeros(1, dtype=np.uint64)
        __compressed_morton_code(grid_coord, axis_bits, output_code)
        return output_code[0]
    else:
        output_code = np.zeros(len(grid_coord), dtype=np.uint64)
        __compressed_morton_code(grid_coord, axis_bits, output_code)
        return output_code


def compressed_morton_decode(morton_code, grid_shape_c_order):
    morton_code = np.asarray(morton_code, np.uint64)
    grid_shape = np.asarray(grid_shape_c_order, np.uint64)

    D = grid_shape.shape[-1]
    axis_bits = np.ceil(np.log2(grid_shape)).astype(np.int8)
    output_shape = np.broadcast_shapes(morton_code.shape, grid_shape.shape[:-1]) + (D,)
    output_grid_coord = np.zeros(output_shape, dtype=np.uint64)
    __compressed_morton_decode(morton_code, axis_bits, output_grid_coord)
    return output_grid_coord


@guvectorize('(),(d)->(d)', nopython=True)
def __compressed_morton_decode(morton_code, axis_bits, grid_coord_c_order):
    D = len(axis_bits)
    curr_axis_pos = np.zeros(D, dtype=np.uint64)
    curr_axis = 0
    
    input_pos = np.uint64(0)
    grid_coord_c_order[:] = np.uint64(0)

    for _ in range(D * 64):
        curr_axis = (curr_axis - 1) % D
        if curr_axis_pos[curr_axis] >= axis_bits[curr_axis]:
            continue

        output_pos = curr_axis_pos[curr_axis]
        grid_coord_c_order[curr_axis] |= ((morton_code >> input_pos) & np.uint64(1)) << output_pos
        input_pos += np.uint64(1)
        curr_axis_pos[curr_axis] += 1



"""
neuroglancer-related utility functions
"""
from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("ngsidekick")
except PackageNotFoundError:
    # Package is not installed
    __version__ = "unknown"

from .gcs import upload_ngstate, upload_ngstates, upload_json, upload_to_bucket, make_bucket_public
from .state_utils import parse_nglink, encode_ngstate, layer_dict, layer_state, download_ngstate
from .annotations.local import local_annotation_json, extract_local_annotations
from .annotations.precomputed import write_precomputed_annotations
from .segmentprops import segment_properties_json, segment_properties_to_dataframe
from .segmentcolors import hex_string_from_segment_id
